@@ -23,6 +23,7 @@ import os
 import re
 import sys
 import numbers
+import operator
 import unittest
 import tempfile
 import functools
@@ -160,6 +161,13 @@ class TestProduct(unittest.TestCase):
     def setUp(self):
         self.product = epr.Product(self.PRODUCT_FILE)
 
+    def test_close(self):
+        self.product.close()
+
+    def test_double_close(self):
+        self.product.close()
+        self.product.close()
+
     def test_file_path_property(self):
         self.assertEqual(self.product.file_path, self.PRODUCT_FILE)
 
@@ -205,7 +213,7 @@ class TestProduct(unittest.TestCase):
 
     else:
 
-        def test_get_dataset_byrtes(self):
+        def test_get_dataset_bytes(self):
             filename = self.DATASET_NAME.encode('UTF-8')
             dataset = self.product.get_dataset(filename)
             self.assertTrue(isinstance(filename, bytes))
@@ -213,7 +221,7 @@ class TestProduct(unittest.TestCase):
 
     def test_datasets(self):
         datasets = [self.product.get_dataset_at(idx)
-                        for idx in range(self.product.get_num_datasets())]
+                    for idx in range(self.product.get_num_datasets())]
         dataset_names = [ds.get_name() for ds in datasets]
         self.assertEqual(dataset_names, self.DATASET_NAMES)
 
@@ -225,7 +233,7 @@ class TestProduct(unittest.TestCase):
         self.assertTrue(isinstance(record, epr.Record))
         product = record.get_field('PRODUCT').get_elem()
         self.assertEqual(product.decode('ascii'),
-                        os.path.basename(self.PRODUCT_FILE))
+                         os.path.basename(self.PRODUCT_FILE))
 
     def test_get_sph(self):
         record = self.product.get_sph()
@@ -260,7 +268,7 @@ class TestProduct(unittest.TestCase):
 
     def test_get_band_id_at_invalid_index(self):
         self.assertRaises(ValueError, self.product.get_band_at,
-                            self.product.get_num_bands())
+                          self.product.get_num_bands())
 
     def test_read_bitmask_raster(self):
         bm_expr = 'l2_flags.LAND AND !l2_flags.CLOUD'
@@ -373,6 +381,16 @@ class TestProductHighLevelAPI(unittest.TestCase):
     def setUp(self):
         self.product = epr.Product(self.PRODUCT_FILE)
 
+    def test_closed(self):
+        self.assertFalse(self.product.closed)
+        self.product.close()
+        self.assertTrue(self.product.closed)
+
+    def test_readonly_closed(self):
+        self.assertFalse(self.product.closed)
+        self.assertRaises(AttributeError,
+                          setattr, self.product, 'closed', True)
+
     def test_get_dataset_names(self):
         self.assertEqual(self.product.get_dataset_names(), self.DATASET_NAMES)
 
@@ -426,6 +444,93 @@ class TestProductHighLevelAPI(unittest.TestCase):
     def test_str_type(self):
         self.assertTrue(isinstance(str(self.product), str))
 
+    def test_contect_manager(self):
+        with epr.open(self.PRODUCT_FILE) as product:
+            self.assertTrue(isinstance(product, epr.Product))
+            self.assertFalse(product.closed)
+            self.assertTrue(str(product))
+
+        self.assertTrue(product.closed)
+
+
+class TestClosedProduct(unittest.TestCase):
+    PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
+    DATASET_NAME = 'Vapour_Content'
+    BAND_NAME = 'water_vapour'
+
+    def setUp(self):
+        self.product = epr.open(self.PRODUCT_FILE)
+        self.product.close()
+
+    def test_properties(self):
+        for name in ('file_path', 'tot_size', 'id_string',
+                     'meris_iodd_version'):
+            self.assertRaises(ValueError, getattr, self.product, name)
+
+    def test_get_get_scene_width(self):
+        self.assertRaises(ValueError, self.product.get_scene_width)
+
+    def test_get_get_scene_height(self):
+        self.assertRaises(ValueError, self.product.get_scene_height)
+
+    def test_get_num_datasets(self):
+        self.assertRaises(ValueError, self.product.get_num_datasets)
+
+    def test_get_num_dsds(self):
+        self.assertRaises(ValueError, self.product.get_num_dsds)
+
+    def test_get_num_bands(self):
+        self.assertRaises(ValueError, self.product.get_num_bands)
+
+    def test_get_mph(self):
+        self.assertRaises(ValueError, self.product.get_mph)
+
+    def test_get_sph(self):
+        self.assertRaises(ValueError, self.product.get_sph)
+
+    def test_get_dataset_at(self):
+        self.assertRaises(ValueError, self.product.get_dataset_at, 0)
+
+    def test_get_dataset(self):
+        self.assertRaises(ValueError, self.product.get_dataset,
+                          self.DATASET_NAME)
+
+    def test_get_dsd_at(self):
+        self.assertRaises(ValueError, self.product.get_dsd_at, 0)
+
+    def test_get_band_id(self):
+        self.assertRaises(ValueError, self.product.get_band, self.BAND_NAME)
+
+    def test_get_band_id_at(self):
+        self.assertRaises(ValueError, self.product.get_band_at, 0)
+
+    def test_read_bitmask_raster(self):
+        bm_expr = 'l2_flags.LAND AND !l2_flags.CLOUD'
+        raster = epr.create_bitmask_raster(12, 10)
+        xoffset = 0
+        yoffset = 0
+
+        self.assertRaises(ValueError, self.product.read_bitmask_raster,
+                          bm_expr, xoffset, yoffset, raster)
+
+    def test_get_dataset_names(self):
+        self.assertRaises(ValueError, self.product.get_dataset_names)
+
+    def test_get_band_names(self):
+        self.assertRaises(ValueError, self.product.get_band_names)
+
+    def test_datasets(self):
+        self.assertRaises(ValueError, self.product.datasets)
+
+    def test_bands(self):
+        self.assertRaises(ValueError, self.product.bands)
+
+    def test_repr(self):
+        self.assertRaises(ValueError, repr, self.product)
+
+    def test_str(self):
+        self.assertRaises(ValueError, str, self.product)
+
 
 class TestDataset(unittest.TestCase):
     PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
@@ -457,10 +562,12 @@ class TestDataset(unittest.TestCase):
         self.assertTrue(isinstance(self.dataset.get_dsd(), epr.DSD))
 
     def test_create_record(self):
-        self.assertTrue(isinstance(self.dataset.create_record(), epr.Record))
+        record = self.dataset.create_record()
+        self.assertTrue(isinstance(record, epr.Record))
 
     def test_read_record(self):
-        self.assertTrue(isinstance(self.dataset.create_record(), epr.Record))
+        record = self.dataset.read_record(0)
+        self.assertTrue(isinstance(record, epr.Record))
 
     def test_read_record_passed(self):
         created_record = self.dataset.create_record()
@@ -519,6 +626,57 @@ class TestDatasetHighLevelAPI(unittest.TestCase):
         self.assertTrue(isinstance(str(self.dataset), str))
 
 
+class TestDatasetOnClosedProduct(unittest.TestCase):
+    PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
+    DATASET_NAME = 'Vapour_Content'
+
+    def setUp(self):
+        self.product = epr.Product(self.PRODUCT_FILE)
+        self.dataset = self.product.get_dataset(self.DATASET_NAME)
+        self.record = self.dataset.create_record()
+        self.product.close()
+
+    def test_product_property(self):
+        self.assertTrue(isinstance(self.dataset.product, epr.Product))
+
+    def test_description_property(self):
+        # @TODO: check
+        self.assertEqual(self.dataset.description, '')
+
+    def test_get_name(self):
+        self.assertRaises(ValueError, self.dataset.get_name)
+
+    def test_get_dsd_name(self):
+        self.assertRaises(ValueError, self.dataset.get_dsd_name)
+
+    def test_get_num_records(self):
+        self.assertRaises(ValueError, self.dataset.get_num_records)
+
+    def test_get_dsd(self):
+        self.assertRaises(ValueError, self.dataset.get_dsd)
+
+    def test_create_record(self):
+        self.assertRaises(ValueError, self.dataset.create_record)
+
+    def test_read_record(self):
+        self.assertRaises(ValueError, self.dataset.read_record, 0)
+
+    def test_read_record_passed(self):
+        self.assertRaises(ValueError, self.dataset.read_record, 0, self.record)
+
+    def test_records(self):
+        self.assertRaises(ValueError, self.dataset.records)
+
+    def test_iter(self):
+        self.assertRaises(ValueError, iter, self.dataset)
+
+    def test_repr(self):
+        self.assertRaises(ValueError, repr, self.dataset)
+
+    def test_str(self):
+        self.assertRaises(ValueError, str, self.dataset)
+
+
 class TestBand(unittest.TestCase):
     PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
     BAND_NAMES = (
@@ -711,7 +869,7 @@ class TestBand(unittest.TestCase):
 
         self.assertEqual(raster1.get_pixel(0, 0), raster2.get_pixel(0, 0))
         self.assertEqual(raster1.get_pixel(width - 1, height - 1),
-                          raster2.get_pixel(width - 1, height - 1))
+                         raster2.get_pixel(width - 1, height - 1))
 
     def test_read_raster_with_invalid_raster(self):
         self.assertRaises(TypeError, self.band.read_raster, 0, 0, 0)
@@ -789,6 +947,80 @@ class TestBandHighLevelAPI(unittest.TestCase):
         self.assertTrue(isinstance(str(band), str))
 
 
+class TestBandOnClosedProduct(unittest.TestCase):
+    PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
+    WIDTH = 12
+    HEIGHT = 10
+
+    def setUp(self):
+        self.product = epr.Product(self.PRODUCT_FILE)
+        self.band = self.product.get_band_at(0)
+        self.raster = self.band.create_compatible_raster(self.WIDTH,
+                                                         self.HEIGHT)
+        self.product.close()
+
+    def test_product_property(self):
+        self.assertTrue(isinstance(self.band.product, epr.Product))
+
+    # @TODO: check
+    #def test_properties(self):
+    #    for name in ('spectr_band_index', 'sample_model', 'data_type',
+    #                 'scaling_method', 'scaling_offset', 'scaling_factor',
+    #                 'bm_expr', 'unit', 'description', 'lines_mirrored'):
+    #        self.assertRaises(ValueError, getattr, self.band, name)
+
+    def test_sample_model_property(self):
+        self.assertEqual(self.band.sample_model, 0)
+
+    def test_data_type_property(self):
+        self.assertEqual(self.band.data_type, 0)
+
+    def test_scaling_method_property(self):
+        self.assertEqual(self.band.scaling_method, epr.E_SMID_LIN)
+
+    def test_scaling_offset_property(self):
+        self.assertEqual(self.band.scaling_offset, 0)
+
+    def test_scaling_factor_property(self):
+        self.assertEqual(self.band.scaling_factor, 0)
+        self.assertTrue(isinstance(self.band.scaling_factor, float))
+
+    def test_bm_expr_property(self):
+        self.assertEqual(self.band.bm_expr, None)
+
+    def test_unit_property(self):
+        self.assertEqual(self.band.unit, None)
+
+    def test_description_property(self):
+        self.assertEqual(self.band.description, None)
+
+    def test_lines_mirrored_property(self):
+        self.assertTrue(isinstance(self.band.lines_mirrored, bool))
+        self.assertEqual(self.band.lines_mirrored, False)
+
+    # END: check
+
+    def test_get_name(self):
+        self.assertRaises(ValueError, self.product.get_band_at, 0)
+
+    def test_create_compatible_raster(self):
+        self.assertRaises(ValueError, self.band.create_compatible_raster,
+                          self.WIDTH, self.HEIGHT)
+
+    def test_read_raster(self):
+        self.assertRaises(ValueError, self.band.read_raster)
+
+    def test_read_as_array(self):
+        self.assertRaises(ValueError, self.band.read_as_array,
+                          self.WIDTH, self.HEIGHT)
+
+    def test_str(self):
+        self.assertRaises(ValueError, str, self.band)
+
+    def test_repr(self):
+        self.assertRaises(ValueError, repr, self.band)
+
+
 class TestCreateRaster(unittest.TestCase):
     RASTER_WIDTH = 400
     RASTER_HEIGHT = 300
@@ -1194,6 +1426,51 @@ class TestMphRecordHighLevelAPI(TestRecordHighLevelAPI):
         self.record = product.get_mph()
 
 
+class TestRecordOnClosedProduct(unittest.TestCase):
+    PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
+    DATASET_NAME = 'Quality_ADS'
+    NUM_FIELD = 21
+    FIELD_NAME = 'perc_water_abs_aero'
+    FIELD_NAMES = TestRecordHighLevelAPI.FIELD_NAMES
+
+    def setUp(self):
+        product = epr.Product(self.PRODUCT_FILE)
+        dataset = product.get_dataset(self.DATASET_NAME)
+        self.record = dataset.read_record(0)
+        product.close()
+
+    # @TODO: check
+    #def test_get_num_fields(self):
+    #    self.assertEqual(self.record.get_num_fields(), self.NUM_FIELD)
+
+    def test_print_(self):
+        self.assertRaises(ValueError, self.record.print_)
+
+    def test_print_element(self):
+        self.assertRaises(ValueError, self.record.print_element, 3, 0)
+
+    def test_get_field(self):
+        self.assertRaises(ValueError, self.record.get_field, self.FIELD_NAME)
+
+    def test_get_field_at(self):
+        self.assertRaises(ValueError, self.record.get_field_at, 0)
+
+    def test_get_field_names(self):
+        self.assertRaises(ValueError, self.record.get_field_names)
+
+    def test_fields(self):
+        self.assertRaises(ValueError, self.record.fields)
+
+    def test_iter(self):
+        self.assertRaises(ValueError, iter, self.record)
+
+    def test_repr(self):
+        self.assertRaises(ValueError, repr, self.record)
+
+    def test_str(self):
+        self.assertRaises(ValueError, str, self.record)
+
+
 class TestField(unittest.TestCase):
     PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
     DATASET_NAME = 'Quality_ADS'
@@ -1365,6 +1642,59 @@ class TestFieldHighLevelAPI2(unittest.TestCase):
     #    self.assertEqual(len(field), len(field.get_elem()))
 
 
+class TestFieldOnClosedProduct(unittest.TestCase):
+    PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
+    DATASET_NAME = 'Quality_ADS'
+    FIELD_NAME = 'perc_water_abs_aero'
+
+    def setUp(self):
+        product = epr.Product(self.PRODUCT_FILE)
+        dataset = product.get_dataset(self.DATASET_NAME)
+        self.record = dataset.read_record(0)
+        self.field = self.record.get_field(self.FIELD_NAME)
+        self.field2 = self.record.get_field_at(4)
+        product.close()
+
+    def test_print_field(self):
+        self.assertRaises(ValueError, self.field.print_)
+
+    def test_get_unit(self):
+        self.assertRaises(ValueError, self.field.get_unit)
+
+    def test_get_description(self):
+        self.assertRaises(ValueError, self.field.get_description)
+
+    def test_get_num_elems(self):
+        self.assertRaises(ValueError, self.field.get_num_elems)
+
+    def test_get_name(self):
+        self.assertRaises(ValueError, self.field.get_name)
+
+    def test_get_type(self):
+        self.assertRaises(ValueError, self.field.get_type)
+
+    def test_get_elem(self):
+        self.assertRaises(ValueError, self.field.get_elem)
+
+    def test_get_elems(self):
+        self.assertRaises(ValueError, self.field.get_elems)
+
+    def test_repr(self):
+        self.assertRaises(ValueError, repr, self.field)
+
+    def test_str(self):
+        self.assertRaises(ValueError, str, self.field)
+
+    def test_eq(self):
+        self.assertRaises(ValueError, operator.eq, self.field, self.field2)
+
+    def test_ne(self):
+        self.assertRaises(ValueError, operator.ne, self.field, self.field2)
+
+    def test_len(self):
+        self.assertRaises(ValueError, len, self.field)
+
+
 class TestDSD(unittest.TestCase):
     PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
     DSD_INDEX = 0
@@ -1454,6 +1784,53 @@ class TestDsdHighLevelAPI(unittest.TestCase):
         self.assertTrue(isinstance(str(self.dsd), str))
 
 
+class TestDSDOnCloserProduct(unittest.TestCase):
+    PRODUCT_FILE = os.path.join(TESTDIR, TEST_PRODUCT)
+    DSD_INDEX = 0
+
+    def setUp(self):
+        self.product = epr.Product(self.PRODUCT_FILE)
+        self.dsd = self.product.get_dsd_at(self.DSD_INDEX)
+        self.dsd2 = self.product.get_dsd_at(self.DSD_INDEX + 1)
+        self.product.close()
+
+    def test_index(self):
+        self.assertRaises(ValueError, getattr, self.dsd, 'index')
+
+    def test_ds_name(self):
+        self.assertRaises(ValueError, getattr, self.dsd, 'ds_name')
+
+    def test_ds_type(self):
+        self.assertRaises(ValueError, getattr, self.dsd, 'ds_type')
+
+    def test_filename(self):
+        self.assertRaises(ValueError, getattr, self.dsd, 'filename')
+
+    def test_ds_offset(self):
+        self.assertRaises(ValueError, getattr, self.dsd, 'ds_offset')
+
+    def test_ds_size(self):
+        self.assertRaises(ValueError, getattr, self.dsd, 'ds_size')
+
+    def test_num_dsr(self):
+        self.assertRaises(ValueError, getattr, self.dsd, 'num_dsr')
+
+    def test_dsr_size(self):
+        self.assertRaises(ValueError, getattr, self.dsd, 'dsr_size')
+
+    def test_eq(self):
+        self.assertRaises(ValueError, operator.eq, self.dsd, self.dsd2)
+
+    def test_ne(self):
+        self.assertRaises(ValueError, operator.ne, self.dsd, self.dsd2)
+
+    def test_repr(self):
+        self.assertRaises(ValueError, repr, self.dsd)
+
+    def test_str(self):
+        self.assertRaises(ValueError, str, self.dsd)
+
+
 class TestDataypeFunctions(unittest.TestCase):
     TYPE_NAMES = {
         epr.E_TID_UNKNOWN: '',
@@ -1553,8 +1930,9 @@ class TestDirectInstantiation(unittest.TestCase):
                 if isinstance(expected_regexp, types.StringTypes):
                     expected_regexp = re.compile(expected_regexp)
                 if not expected_regexp.search(str(exc_value)):
-                    raise self.failureException('"%s" does not match "%s"' %
-                             (expected_regexp.pattern, str(exc_value)))
+                    raise self.failureException(
+                        '"%s" does not match "%s"' % (expected_regexp.pattern,
+                                                      str(exc_value)))
             else:
                 if hasattr(expected_exception, '__name__'):
                     excName = expected_exception.__name__
